#define _CRT_SECURE_NO_DEPRECATE

#include <windows.h>
#include "filewriter_immediate.h"


const int		cImmediateBufferSize			= 16384;

unsigned char*	FileWriterImmediate::mBuffer	= NULL;


FileWriterImmediate* FileWriterImmediate::GetFileWriter(const std::string& inName, long long inFileSize)
{
	if (!mBuffer)
		mBuffer = new unsigned char[cImmediateBufferSize];
	return new FileWriterImmediate(inName, inFileSize);
}


void FileWriterImmediate::Finalize()
{
	delete mBuffer;
	mBuffer = NULL;
}


FileWriterImmediate::FileWriterImmediate(const std::string& inName, long long inFileSize) :
	FileWriter(inName, inFileSize),
	mHandle(INVALID_HANDLE_VALUE)
{
}


bool FileWriterImmediate::Open(bool inWriteable, int inModTime)
{
	mHandle = CreateFile(inWriteable, false, inModTime);
	return mHandle != INVALID_HANDLE_VALUE;
}


bool FileWriterImmediate::Close()
{
	return CloseFile(mHandle);
}


void FileWriterImmediate::GetWriteBuffer(unsigned char*& outBuf, int& outLength)
{
	outBuf = mBuffer;
	outLength = cImmediateBufferSize;
}


bool FileWriterImmediate::ReleaseWriteBuffer(int inRemainingLength)
{
	int length = cImmediateBufferSize - inRemainingLength;
	if (length > 0)
	{
		DWORD written;
		if (::WriteFile(mHandle, mBuffer, length, &written, NULL) == 0 || written != length)
			return false;
		if (mWriteCallback)
			mWriteCallback(written);
	}
	return true;
}


void FileWriterImmediate::HandleAbort()
{
	CloseFile(mHandle);
	// Aborted before file actions could be finished, so delete
	DeleteFile();
}


bool FileWriterImmediate::ChModTime(int inModTime)
{
	return FileWriter::ChModTime(mHandle, inModTime);
}

